@@ -43,6 +43,7 @@
 #include <glib/gi18n.h>
 #include <errno.h>
 #include <string.h>
+#include <sys/ioctl.h>
 #include <sys/types.h>
 #include <unistd.h>
 
@@ -53,6 +54,60 @@
 
 #define N_WAKEUP_EVENTS 149
 
+/* Identifiers for the various tracepoints we might watch for */
+enum SpTracepoint {
+  DRM_VBLANK,
+  DRM_I915_BEGIN,
+  DRM_I915_END,
+};
+
+typedef struct {
+  enum SpTracepoint tp;
+  const char       *path;
+  const char      **fields;
+} SpOptionalTracepoint;
+
+/* Global list of the optional tracepoints we might want to watch. */
+static const SpOptionalTracepoint optional_tracepoints[] = {
+  /* This event fires just after the vblank IRQ handler starts.
+   *
+   * Note that on many platforms when nothing is waiting for vblank
+   * (no pageflips have happened recently, no rendering is
+   * synchronizing to vblank), the vblank IRQ will get masked off and
+   * the event won't show up in the timeline.
+   *
+   * Also note that when we're in watch-a-single-process mode, we
+   * won't get the event since it comes in on an IRQ handler, not for
+   * our pid.
+   */
+  { DRM_VBLANK, "drm/drm_vblank_event",
+    (const char *[]){ "crtc", "seq", NULL } },
+
+  /* I915 GPU execution.
+   *
+   * These are the wrong events to be watching.  We need to use the
+   * ones under CONFIG_DRM_I915_LOW_LEVEL_TRACEPOINTS instead.
+   */
+#if 0
+  { DRM_I915_BEGIN, "i915/i915_gem_request_add",
+    (const char *[]){ "ctx", "ring", "seqno", NULL } },
+  { DRM_I915_END, "i915/i915_gem_request_retire",
+    (const char *[]){ "ctx", "ring", "seqno", NULL } },
+#endif
+};
+
+/* Struct describing tracepoint events.
+ *
+ * This should be extended with some sort of union for the describing
+ * the locations of the relevant fields within the _RAW section of the
+ * struct perf_event, so we can pick out things like the vblank CRTC
+ * number and MSC.
+ */
+typedef struct {
+  enum SpTracepoint tp;
+  gsize field_offsets[0];
+} SpTracepointDesc;
+
 struct _SpPerfSource
 {
   GObject          parent_instance;
@@ -61,6 +116,9 @@ struct _SpPerfSource
   SpPerfCounter   *counter;
   GHashTable      *pids;
 
+  /* Mapping from perf sample identifiers to SpTracepointDesc. */
+  GHashTable      *tracepoint_event_ids;
+
   guint            running : 1;
   guint            is_ready : 1;
 };
@@ -93,6 +151,7 @@ sp_perf_source_finalize (GObject *object)
   g_clear_pointer (&self->writer, sp_capture_writer_unref);
   g_clear_pointer (&self->counter, sp_perf_counter_unref);
   g_clear_pointer (&self->pids, g_hash_table_unref);
+  g_clear_pointer (&self->tracepoint_event_ids, g_hash_table_unref);
 
   G_OBJECT_CLASS (sp_perf_source_parent_class)->finalize (object);
 }
@@ -116,6 +175,7 @@ static void
 sp_perf_source_init (SpPerfSource *self)
 {
   self->pids = g_hash_table_new (NULL, NULL);
+  self->tracepoint_event_ids = g_hash_table_new (NULL, NULL);
 }
 
 static gboolean
@@ -129,9 +189,67 @@ do_emit_exited (gpointer data)
 }
 
 static void
-sp_perf_source_handle_sample (SpPerfSource                   *self,
-                              gint                            cpu,
-                              const SpPerfCounterEventSample *sample)
+sp_perf_source_handle_tracepoint (SpPerfSource                       *self,
+                                  gint                                cpu,
+                                  const SpPerfCounterEventTracepoint *sample,
+                                  SpTracepointDesc                   *tp_desc)
+{
+  gchar *message = NULL;
+
+  /* Note that field_offsets[] correspond to the
+   * SpOptionalTracepoint->fields[] strings.  Yes, this is gross.
+   */
+  switch (tp_desc->tp)
+    {
+    case DRM_VBLANK:
+      message = g_strdup_printf ("crtc=%d, seq=%u",
+                                 *(gint *)(sample->raw +
+                                           tp_desc->field_offsets[0]),
+                                 *(guint *)(sample->raw +
+                                            tp_desc->field_offsets[1]));
+
+      sp_capture_writer_add_mark (self->writer,
+                                  sample->time,
+                                  cpu,
+                                  sample->pid,
+                                  0,
+                                  "drm",
+                                  "vblank",
+                                  message);
+      break;
+
+    case DRM_I915_BEGIN:
+    case DRM_I915_END:
+      message = g_strdup_printf ("ctx=%u, ring=%u, seqno=%u",
+                                 *(guint *)(sample->raw +
+                                            tp_desc->field_offsets[0]),
+                                 *(guint *)(sample->raw +
+                                            tp_desc->field_offsets[1]),
+                                 *(guint *)(sample->raw +
+                                            tp_desc->field_offsets[2]));
+
+      sp_capture_writer_add_mark (self->writer,
+                                  sample->time,
+                                  cpu,
+                                  sample->pid,
+                                  0,
+                                  "drm",
+                                  (tp_desc->tp == DRM_I915_BEGIN ?
+                                   "i915 gpu begin" : "i915 gpu end"),
+                                  message);
+      break;
+
+    default:
+      break;
+    }
+
+  g_free (message);
+}
+
+static void
+sp_perf_source_handle_callchain (SpPerfSource                      *self,
+                                 gint                               cpu,
+                                 const SpPerfCounterEventCallchain *sample)
 {
   const guint64 *ips;
   gint n_ips;
@@ -185,6 +303,7 @@ sp_perf_source_handle_event (SpPerfCounterEvent *event,
                              gpointer            user_data)
 {
   SpPerfSource *self = user_data;
+  SpTracepointDesc *tp_desc;
   gsize offset;
   gint64 time;
 
@@ -266,7 +385,21 @@ sp_perf_source_handle_event (SpPerfCounterEvent *event,
       break;
 
     case PERF_RECORD_SAMPLE:
-      sp_perf_source_handle_sample (self, cpu, &event->sample);
+      /* We don't capture IPs with tracepoints, and get _RAW data
+       * instead.  Handle them separately.
+       */
+      g_assert (&event->callchain.identifier == &event->tracepoint.identifier);
+      tp_desc = g_hash_table_lookup (self->tracepoint_event_ids,
+                                     GINT_TO_POINTER (event->callchain.identifier));
+      if (tp_desc)
+        {
+          sp_perf_source_handle_tracepoint (self, cpu, &event->tracepoint,
+                                            tp_desc);
+        }
+      else
+        {
+          sp_perf_source_handle_callchain (self, cpu, &event->callchain);
+        }
       break;
 
     case PERF_RECORD_THROTTLE:
@@ -276,6 +409,187 @@ sp_perf_source_handle_event (SpPerfCounterEvent *event,
     }
 }
 
+static gboolean
+sp_perf_get_tracepoint_config (const char *path, gint64 *config)
+{
+  gchar *filename = NULL;
+  gchar *contents;
+  size_t len;
+
+  filename = g_strdup_printf ("/sys/kernel/debug/tracing/events/%s/id", path);
+  if (!filename)
+    return FALSE;
+
+  if (!g_file_get_contents (filename, &contents, &len, NULL))
+    {
+      g_free (filename);
+      return FALSE;
+    }
+
+  g_free(filename);
+
+  *config = strtoull(contents, NULL, 0);
+
+  g_free (contents);
+
+  return TRUE;
+}
+
+static gboolean
+sp_perf_get_tracepoint_fields (SpTracepointDesc           *tp_desc,
+                               const SpOptionalTracepoint *optional_tp,
+                               GError                    **error)
+{
+  gchar *filename = NULL;
+  gchar *contents;
+  size_t len;
+  gint i;
+  filename = g_strdup_printf ("/sys/kernel/debug/tracing/events/%s/format",
+                              optional_tp->path);
+  if (!filename)
+    return FALSE;
+
+  if (!g_file_get_contents (filename, &contents, &len, NULL))
+    {
+      g_free (filename);
+      return FALSE;
+    }
+
+  g_free (filename);
+
+  /* Look up our fields.  Some example strings:
+   *
+   *  field:unsigned short common_type;	offset:0;	size:2;	signed:0;
+   *	field:int crtc;	offset:8;	size:4;	signed:1;
+   *  field:unsigned int seq;	offset:12;	size:4;	signed:0;
+   */
+  for (i = 0; optional_tp->fields[i] != NULL; i++)
+    {
+      gchar *pattern = g_strdup_printf ("%s;\toffset:", optional_tp->fields[i]);
+      gchar *match;
+      gint64 offset;
+
+      match = strstr (contents, pattern);
+      if (!match)
+        {
+          g_set_error (error,
+                       G_IO_ERROR,
+                       G_IO_ERROR_FAILED,
+                       _("Sysprof failed to find field “%s”."),
+                       optional_tp->fields[i]);
+          g_free (contents);
+          return FALSE;
+        }
+
+      offset = g_ascii_strtoll (match + strlen (pattern),
+                                NULL, 0);
+      if (offset == G_MININT64 && errno != 0)
+        {
+          g_set_error (error,
+                       G_IO_ERROR,
+                       G_IO_ERROR_FAILED,
+                       _("Sysprof failed to parse offset for “%s”."),
+                       optional_tp->fields[i]);
+          g_free (contents);
+          return FALSE;
+        }
+
+      tp_desc->field_offsets[i] = offset;
+      g_free (pattern);
+    }
+
+  g_free (contents);
+
+  return TRUE;
+}
+
+/* Adds a perf tracepoint event, if it's available.
+ *
+ * These are kernel tracepoints that we want to include in our capture
+ * when present, but may be kernel version or driver-specific.
+ */
+static void
+sp_perf_source_add_optional_tracepoint (SpPerfSource                *self,
+                                        GPid                        pid,
+                                        gint                        cpu,
+                                        const SpOptionalTracepoint *optional_tracepoint,
+                                        GError                     **error)
+{
+  struct perf_event_attr attr = { 0 };
+  SpTracepointDesc *tp_desc;
+  gulong flags = 0;
+  gint fd;
+  gint64 config;
+  gint64 id;
+  int ret;
+  gint num_fields;
+
+  if (!sp_perf_get_tracepoint_config(optional_tracepoint->path, &config))
+    return;
+
+  attr.type = PERF_TYPE_TRACEPOINT;
+  attr.sample_type = PERF_SAMPLE_RAW
+                   | PERF_SAMPLE_IP
+                   | PERF_SAMPLE_TID
+                   | PERF_SAMPLE_IDENTIFIER
+                   | PERF_SAMPLE_RAW
+                   | PERF_SAMPLE_TIME;
+  attr.config = config;
+  attr.sample_period = 1;
+
+#ifdef HAVE_PERF_CLOCKID
+  attr.clockid = sp_clock;
+  attr.use_clockid = 1;
+#endif
+
+  attr.size = sizeof attr;
+
+  fd = sp_perf_counter_open (self->counter, &attr, pid, cpu, -1, flags);
+
+  ret = ioctl (fd, PERF_EVENT_IOC_ID, &id);
+  if (ret != 0)
+    {
+      g_set_error (error,
+                   G_IO_ERROR,
+                   G_IO_ERROR_FAILED,
+                   _("Sysprof failed to get perf_event ID."));
+      close(fd);
+      return;
+    }
+
+  /* The fields list is NULL-terminated, count how many are there. */
+  for (num_fields = 0; optional_tracepoint->fields[num_fields]; num_fields++)
+    ;
+
+  tp_desc = g_malloc (sizeof (*tp_desc) +
+                      sizeof(*tp_desc->field_offsets) * num_fields);
+  if (!tp_desc)
+    {
+      close(fd);
+      return;
+    }
+
+  tp_desc->tp = optional_tracepoint->tp;
+
+  if (!sp_perf_get_tracepoint_fields (tp_desc, optional_tracepoint, error))
+    {
+      free(tp_desc);
+      close(fd);
+      return;
+    }
+
+  /* Here's where we should inspect the /format file to determine how
+   * to pick fields out of the _RAW data.
+   */
+
+  /* We're truncating the event ID from 64b to 32 to fit in the hash.
+   * The event IDs start from 0 at boot, so meh.
+   */
+  g_assert (id <= 0xffffffff);
+  g_hash_table_insert (self->tracepoint_event_ids,
+                       GINT_TO_POINTER (id), tp_desc);
+}
+
 static gboolean
 sp_perf_source_start_pid (SpPerfSource  *self,
                           GPid           pid,
@@ -286,11 +600,13 @@ sp_perf_source_start_pid (SpPerfSource  *self,
   gint ncpu = g_get_num_processors ();
   gint cpu = 0;
   gint fd;
+  gint i;
 
   g_assert (SP_IS_PERF_SOURCE (self));
 
   attr.sample_type = PERF_SAMPLE_IP
                    | PERF_SAMPLE_TID
+                   | PERF_SAMPLE_IDENTIFIER
                    | PERF_SAMPLE_CALLCHAIN
                    | PERF_SAMPLE_TIME;
   attr.wakeup_events = N_WAKEUP_EVENTS;
@@ -314,6 +630,9 @@ sp_perf_source_start_pid (SpPerfSource  *self,
       cpu = -1;
     }
 
+  /* Perf won't let us capture on all CPUs on all pids, so we have to
+   * loop over CPUs if we're not just watching a single pid.
+   */
   for (; cpu < ncpu; cpu++)
     {
       attr.type = PERF_TYPE_HARDWARE;
@@ -355,6 +674,13 @@ sp_perf_source_start_pid (SpPerfSource  *self,
               return FALSE;
             }
         }
+
+      for (i = 0; i < G_N_ELEMENTS(optional_tracepoints); i++)
+        {
+          sp_perf_source_add_optional_tracepoint (self, pid, cpu,
+                                                  &optional_tracepoints[i],
+                                                  error);
+        }
     }
 
   return TRUE;
